#ifndef _NEURAL_NETWORK_H
#define _NEURAL_NETWORK_H 1

#include<iostream>
#include<vector>
#include<cmath>
#include"rand.h"
#include<cfloat>
#include<map>

using namespace std;

extern const long double epsilon;
bool isToPredict(const vector<long>&datum,long pos);

struct PredictorNetwork{
	long alph; //alphabet size
	int n; //Network size
	int conn; //Network connectivity (edges per unit)
	virtual int suggestedConn()=0;// suggestion of connectivity to balance out various terms in the algorithmic cost of training. Needs to have other parameters already set
	int nedges; //usually n*conn
	vector<int> edgesrc,edgedest; //for an integer e in [0;nedges), edgesrc[e] and edgedest[e] are the source and target of edge e
	vector<vector<pair<int,int> > > edgesto,edgesfrom;//edge reverse index: edgesto[j] is a vector containing all pairs (i,e) where e is an edge from i to j, and likewise edgesfrom[i] is a vector containing all pairs (j,e) with e an edge from i to j

	vector<vector< long double> > pred;  //pred[t][x] = predicted probability to have symbol x at time t
	vector<long double> readfreq,predfreq; //average symbol frequencies in the data

	PredictorNetwork();
	virtual ~PredictorNetwork();
	virtual void buildCompleteGraph(int an);
	virtual void buildErdosRenyi_loops(int an,int aconn);
	virtual void makeEdgeIndex();
	virtual void Setup(long maxdatalength); //builds the arrays to store values
	virtual void getFreqs(const vector<long>& datum); //get symbol frequencies
	virtual void initWeights()=0;
	virtual void setToStartAct()=0;
	virtual void computeNextAct(long t,long symbol_read)=0;//fills act[t] given act[t-1] and last symbol read
	virtual void computePred(long t)=0;//fills pred[t][x] from act[t]
	virtual long double computePred(long t,long x)=0;//fills and returns only pred[t][x] (useful in case pred[t][x] can be computed for each x individually)
	virtual pair<long double,long> computeAct(const vector<long>& somedatum,bool regul=false); //returns base-2 log-likelihood and number of errors
	virtual void saveParams();
	virtual void restoreParams();
	virtual void saveWParams()=0;
	virtual void restoreWParams()=0;
	virtual void saveTauParams()=0;
	virtual void restoreTauParams()=0;
	virtual bool sample(unsigned long length,vector<long>& sampleresult);//get a sample from current params. Overwrites activity values.
};

struct NeuralNetwork:public virtual PredictorNetwork{
public:

	vector<vector<long double> > act;  //act[t][i] is the activity of unit i at time t
	vector<vector<long double> > V;  // act[t][i] = actfunc (V[t][i])

	virtual int suggestedConn()=0;// suggestion of connectivity to balance out various terms in the algorithmic cost of training. Needs to have other parameters already set
	virtual long double actfunc(long double energy)=0;//activation function
	virtual long double deractfunc(long double activity)=0;//derivative of the activation function, EXPRESSED AS A FUNCTION OF ACTIVITY (i.e., actfunc'(invactfunc), e.g. 1-act^2 for tanh
	virtual long double invactfunc(long double activity)=0;
	virtual void Setup(long maxdatalength); //builds the arrays to store values
	virtual void computeWGrad(const vector<long>& datum)=0;
	virtual void computeTauGrad(const vector<long>& datum)=0;
	virtual void applyWGrad(long double rate)=0;
	virtual void applyTauGrad(long double rate)=0;
	virtual void saveWParams()=0;
	virtual void restoreWParams()=0;
	virtual void saveTauParams()=0;
	virtual void restoreTauParams()=0;
};


struct NN_SoftMaxOutput:public virtual NeuralNetwork{
	typedef enum {Unset,BackpropW,DiagInvHessW,QDInvHessW,QDInvHessWwithRenorm} WGradType;
	WGradType wGradType;
	vector<vector<long double> > w,wgrad,prev_w;  //w[x][i] writing weight for symbol x at unit i
	vector<long double > wbias,wbiasgrad,prev_wbias; //wbias[x] writing bias for symbol x = writing weight from always-activated unit
	vector<long double> energy;
	vector<long double> hess00; // hess00[x] = diag entry of the Hessian corresponding to wbias[x]
	vector<vector<long double> > hessii;// hessii[x][i] = diag entry of the Hessian corresponding to w[x][i]
	vector<vector<long double> > hess0i;// hess0i[x][i] = entry of the Hessian corresponding to wbias[x] and w[x][i]
	void computePred(long t);//fills pred[t][x]
	long double computePred(long t,long x);//fills pred[t][x]
	void fillProbFromEnergy(const vector<long double>& someenergy,vector<long double>& prob);
	void computeWDiff(const vector<long>& datum);
	void computeWGrad(const vector<long>& datum);
	void computeWMetric(const vector<long>& datum,bool includecrossterms);
	long double computeFisherMetric(const vector<long>& datum,const vector<vector<long double> >& dw,const vector<long double>& dwbias);
	void applyWGrad(long double rate);
	void saveWParams();void restoreWParams();
	virtual void Setup(long maxDatumLength);
};

struct NN_tanhAct:public virtual NeuralNetwork{
	virtual long double actfunc(long double energy);
	virtual long double deractfunc(long double theact);
	virtual long double invactfunc(long double theact);
};

struct NN_logisticAct:public virtual NeuralNetwork{
	virtual long double actfunc(long double energy);
	virtual long double deractfunc(long double theact);
	virtual long double invactfunc(long double theact);
};

struct NN_normcdfAct:public virtual NeuralNetwork{
	virtual long double actfunc(long double energy);
	virtual long double deractfunc(long double theact);//NOT DEFINED, will return error
	virtual long double deractfuncV(long double energy);
	virtual long double invactfunc(long double theact);
};

#endif
